import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:get/get.dart';
import 'package:wowit/firebaseService.dart';

abstract class BaseController extends GetxController {
  String _collection;
  Map<String, List<dynamic>> _dictCachedData = {};
  Rx<List<dynamic>> _rxData = Rx<List<dynamic>>();

  //condition
  Map<String, Object> _conditions;
  Map<String, Object> _orderBy;
  int _limit = -1;
  bool _usingDefaultCondition = true;
  String _keyCachedData;
  bool _forceReload = true;

  List<S> getData<S>({String keyCachedData}) {
    bool isCached = keyCachedData != null && keyCachedData != "";

    if (isCached && _dictCachedData.containsKey(keyCachedData))
      return _dictCachedData[keyCachedData];
    else
      return _rxData.value;
  }

  BaseController(
    this._collection,
  );

  @override
  void update([List<Object> ids, bool condition = true]) {
    // TODO: implement update
    super.update(ids, condition);
    print(this.runtimeType.toString() + ' - BaseController on update');
  }

  @override
  void onClose() {
    // TODO: implement onClose
    super.onClose();
    print(this.runtimeType.toString() + ' - BaseController on onClose');
  }

  @override
  void onReady() {
    // TODO: implement onReady
    super.onReady();
    print(this.runtimeType.toString() + ' - BaseController on onReady');
  }

  @override
  void onInit() {
    // TODO: implement onInit
    super.onInit();
    print(this.runtimeType.toString() +
        ' - BaseController on onInit with condition');
    _rxData.bindStream(onBindStream(
      conditions: this._conditions,
      orderBy: this._orderBy,
      limit: this._limit,
      usingDefaultCondition: this._usingDefaultCondition,
    keyCachedData: this._keyCachedData, forceReload: this._forceReload));
  }

  void setUpBindStream(
      {Map<String, Object> conditions,
      Map<String, Object> orderBy,
      int limit = -1,
      bool usingDefaultCondition = true,
      String keyCachedData,
      bool forceReload = true}) {
    _conditions = conditions;
    _orderBy = orderBy;
    _limit = limit;
    _usingDefaultCondition = usingDefaultCondition;
    _keyCachedData = keyCachedData;
    _forceReload = forceReload;
  }

  Stream<List<S>> onBindStream<S>(
      {Map<String, Object> conditions,
      Map<String, Object> orderBy,
      int limit = -1,
      bool usingDefaultCondition = true,
      String keyCachedData,
      bool forceReload = true}) {
    if (this._collection == '') {
      throw ("[ERROR] Collection is empty");
    }
    // TODO: implement onBindStream
    if (usingDefaultCondition != null && usingDefaultCondition) {
      if (conditions == null) conditions = {};
      if (!conditions.containsKey('statusData')) conditions['statusData'] = 3;
      if (!conditions.containsKey('visible')) conditions['visible'] = true;
    }
    bool isCached = keyCachedData != null && keyCachedData != "";
    if (isCached && !forceReload) //get data in cache
    {
      if (_dictCachedData.containsKey(keyCachedData)) {
        //return onBeforeReturnResult(keyCachedData[keyCachedData]);// not support
      }
    }
    print(this.runtimeType.toString() +
        ' - BaseController on onBindStream - conditions: ' +
        conditions.toString());
    if (isCached) _dictCachedData[keyCachedData] = [];
    return FirebaseService.getQuery(this._collection,
            conditions: conditions, orderBy: orderBy, limit: limit)
        .snapshots()
        .map((QuerySnapshot query) {
      List<S> retVal = [];
      query.docs.forEach((element) {
        dynamic item = toModel(element);
        retVal.add(item);
        if (isCached) _dictCachedData[keyCachedData].add(item);
      });
      return onBeforeReturnResult(retVal);
    });
  }

  List<S> onBeforeReturnResult<S>(List<S> retVal) {
    return retVal;
  }

  //abstract methods
  S toModel<S>(DocumentSnapshot element);
}
